import threading
import transaction
import zope.interface
import transaction.interfaces


class CelerySession(threading.local):
    """Thread local session of data to be sent to Celery."""

    def __init__(self):
        self.tasks = []
        self._needs_to_join = True

    def add_call(self, method, *args, **kw):
        self._join_transaction()
        self.tasks.append((method, args, kw))

    def reset(self):
        self.tasks = []
        self._needs_to_join = True

    def _join_transaction(self):
        if not self._needs_to_join:
            return
        dm = CeleryDataManager(self)
        transaction.get().join(dm)
        self._needs_to_join = False

    def _flush(self):
        for method, args, kw in self.tasks:
            method(*args, **kw)
        self.reset()

    def __len__(self):
        """Number of tasks in the session."""
        return len(self.tasks)


celery_session = CelerySession()


@zope.interface.implementer(transaction.interfaces.IDataManager)
class CeleryDataManager:
    """DataManager embedding the access to celery into the transaction."""

    transaction_manager = None

    def __init__(self, session):
        self.session = session

    def abort(self, transaction):
        self.session.reset()

    def tpc_begin(self, transaction):
        pass

    def commit(self, transaction):
        pass

    tpc_abort = abort

    def tpc_vote(self, transaction):
        self.session._flush()

    def tpc_finish(self, transaction):
        pass

    def sortKey(self):
        # Sort last, so that sending to celery is done after all other
        # DataManagers signalled an okay.
        return "~z3c.celery"

    def __repr__(self):
        """Custom repr."""
        return '<{0.__module__}.{0.__name__} for {1}, {2}>'.format(
            self.__class__, transaction.get(), self.session)



from ..loader import ZopeLoader
from .shared_tasks import get_principal_title_task
from unittest import mock
from zope.principalregistry.principalregistry import principalRegistry
import celery.signals
import contextlib
import logging
import plone.testing.zca
import pytest
import tempfile
import z3c.celery
import zope.app.appsetup.appsetup
import zope.security.management


LOGGING_TEMPLATE = """
[loggers]
keys = root, root

[handlers]
keys = logfile

[formatters]
keys = taskformatter

[logger_root]
level = DEBUG
handlers = logfile
qualname = root

[handler_logfile]
class = FileHandler
formatter = taskformatter
args = ('{filename}',)

[formatter_taskformatter]
class = z3c.celery.logging.TaskFormatter
format = task_id: %(task_id)s name: %(task_name)s %(message)s
"""


def test_loader__ZopeLoader__on_worker_init__1__cov(
        interaction, eager_celery_app, zcml):
    """It connects a signal for logging setup if LOGGING_INI is present in

    the configuration.

    As it is hard to collect coverage for sub-processes we use this test for
    coverage only.
    """
    @z3c.celery.task
    def simple_log():
        """Just log something."""
        log = logging.getLogger(__name__)
        log.debug('Hello Log!')

    configure_zope = 'z3c.celery.celery.TransactionAwareTask.configure_zope'
    with tempfile.NamedTemporaryFile(delete=False) as logging_ini, \
            tempfile.NamedTemporaryFile(delete=False) as logfile:
        logging_ini.write(
            LOGGING_TEMPLATE.format(filename=logfile.name).encode('utf-8'))
        logging_ini.flush()
        eager_celery_app.conf['LOGGING_INI'] = logging_ini.name
        loader = ZopeLoader(app=eager_celery_app)
        loader.on_worker_init()
        celery.signals.setup_logging.send(sender=None)

        with mock.patch(configure_zope):
            zope.security.management.endInteraction()
            simple_log(
                _run_asynchronously_=True, _principal_id_='example.user')

        logfile.seek(0)
        log_result = logfile.read().decode('utf-8')

        assert ('task_id: None name: z3c.celery.tests.test_loader.'
                'simple_log Hello Log!\n' in log_result)


def test_loader__ZopeLoader__on_worker_process_init__1__cov(
        interaction, eager_celery_app):
    """It raises a ValueError if there is no `ZOPE_CONF` in the configuration.

    As it is hard to collect coverage for sub-processes we use this test for
    coverage only.
    """
    eager_celery_app.conf['ZOPE_CONF'] = None

    loader = ZopeLoader(app=eager_celery_app)
    with pytest.raises(ValueError) as err:
        loader.on_worker_process_init()
    assert (
        'Celery setting ZOPE_CONF not set, check celery worker config.' ==
        str(err.value))


@contextlib.contextmanager
def zope_loader(app):
    plone.testing.zca.pushGlobalRegistry()
    try:
        app.loader.on_worker_process_init()
        yield
    finally:
        app.loader.on_worker_shutdown()
        plone.testing.zca.popGlobalRegistry()
        principalRegistry._clear()
        zope.app.appsetup.appsetup.reset()


def test_loader__ZopeLoader__1__cov(eager_celery_app):
    """It sets up and tears down Zope on worker init and teardown.

    As it is hard to collect coverage for sub-processes we use this test for
    coverage only.
    """
    with zope_loader(eager_celery_app):
        assert ('Ben Utzer' ==
                principalRegistry.getPrincipal('example.user').title)


def test_loader__ZopeLoader__2__cov(eager_celery_app):
    """It provides the Zope environment to the task."

    As it is hard to collect coverage for sub-processes we use this test for
    coverage only.
    """
    with zope_loader(eager_celery_app):
        assert "Ben Utzer" == get_principal_title_task(
            _run_asynchronously_=True, _principal_id_='example.user')



from ..session import celery_session, CeleryDataManager


def test_session__CeleryDataManager____repr____1():
    """It has a custom repr."""
    dm = CeleryDataManager(celery_session)
    assert repr(dm).startswith(
        '<z3c.celery.session.CeleryDataManager for <transaction')



from celery._state import get_current_task
import zope.exceptions.log


class TaskId:

    def format(self, record):
        task = get_current_task()
        if task:
            record.__dict__.update(
                task_id=task.request.id,
                task_name=task.name)
        else:
            record.__dict__.setdefault('task_name', '')
            record.__dict__.setdefault('task_id', '')
        return super().format(record)


class TaskFormatter(TaskId, zope.exceptions.log.Formatter):
    """Provides `task_id` and `task_name` variables for the log format.

    We want to have a general formatter so we want to get rid of '???' which
    are rendered by celery.app.log.TaskFormatter. Also we inherit from
    zope.exceptions to support `__traceback_info__`.
    """


try:
    import pythonjsonlogger.jsonlogger
except ImportError:
    pass
else:
    class JsonFormatter(TaskId, pythonjsonlogger.jsonlogger.JsonFormatter):
        pass



from .loader import ZopeLoader
from .session import celery_session
from celery._state import _task_stack
from celery.utils.serialization import raise_with_context
import ZODB.POSException
import celery
import celery.exceptions
import celery.utils
import contextlib
import json
import logging
import random
import socket
import time
import transaction
import zope.app.publication.zopepublication
import zope.authentication.interfaces
import zope.component
import zope.component.hooks
import zope.publisher.browser
import zope.security.management


log = logging.getLogger(__name__)


class HandleAfterAbort(RuntimeError):
    """Exception whose callback is executed after ``transaction.abort()``."""

    def __init__(self, callback, *args, **kwargs):
        self.message = kwargs.pop('message', u'')
        # Conform to BaseException API so it works with celery's serialization
        # (before 4.1.1 it just worked, but that was rather accidental)
        if isinstance(callback, str):
            self.message = callback
        if isinstance(self.message, bytes):
            self.message = self.message.decode('utf-8')

        super().__init__(self.message)

        self.callback = callback
        self.c_args = args
        self.c_kwargs = kwargs

    def __call__(self):
        self.callback(*self.c_args, **self.c_kwargs)

    def __str__(self):
        return self.message


class Abort(HandleAfterAbort):
    """Exception to signal successful task completion, but transaction should
    be aborted instead of commited.
    """


class Retry(celery.exceptions.Retry, HandleAfterAbort):
    """With cooperation from TransactionAwareTask.retry(), this moves the
    actual re-queueing of the task into the proper "error handling"
    transaction phase.
    """

    def __init__(self, *args, **kw):
        self.signature = kw.pop('signature', None)
        celery.exceptions.Retry.__init__(self, *args, **kw)

    def __call__(self):
        try:
            self.signature.apply_async()
            log.warning('Task submitted for retry in %s seconds.', self.when)
        except Exception as exc:
            log.error('Task retry failed', exc_info=True)
            raise celery.exceptions.Reject(exc, requeue=False)


def get_principal(principal_id):
    """Return the principal to the principal_id."""
    auth = zope.component.getUtility(
        zope.authentication.interfaces.IAuthentication)
    return auth.getPrincipal(principal_id)


def login_principal(principal, task_name=None):
    """Start an interaction with `principal`."""
    request = zope.publisher.browser.TestRequest(
        environ={'SERVER_URL': 'http://%s' % socket.getfqdn()})
    if task_name:
        # I'd rather set PATH_INFO, but that doesn't influence getURL().
        request._traversed_names = ['celery', task_name]
    request.setPrincipal(principal)
    zope.security.management.newInteraction(request)


class TransactionAwareTask(celery.Task):
    """Wrap every Task execution in a transaction begin/commit/abort.

    (Code inspired by gocept.runner.runner.MainLoop)
    """

    abstract = True  # Base class. Don't register as an executable task.

    def __call__(self, *args, **kw):
        """Run the task.

        Parameters:
            _run_asynchronously_ ... if `True` run the task its own transaction
                                     context, otherwise run it in inline
                                     (optional, default: `False` to be able to
                                      run tasks easily inline)
            _principal_id_ ... run asynchronous task as this user, ignored if
                               running synchronously (optional)

        Returns whatever the task returns itself.

        """
        run_asynchronously = kw.pop('_run_asynchronously_', False)
        principal_id = kw.pop('_principal_id_', None)
        # BBB This was removed in 1.2.0 but there still might be (scheduled)
        # tasks that transmit this argument, so we need to remove it.
        kw.pop('_task_id_', None)

        is_eager = self.app.conf['task_always_eager']
        if is_eager:
            # This is the only part of celery.Task.__call__ that actually is
            # relevant -- since in non-eager mode, it's not called at all:
            # celery.app.trace.build_tracer() says, "if the task doesn't define
            # a custom __call__ method we optimize it away by simply calling
            # the run method directly".
            # (Note that the push_request() call in __call__ would be actively
            # harmful in non-eager mode, since it hides the actual request that
            # was set by app.trace; but as it's not called, it's not an issue.)
            _task_stack.push(self)
        try:
            if run_asynchronously:
                result = self.run_in_worker(principal_id, args, kw)
            else:
                result = self.run_in_same_process(args, kw)
        finally:
            if is_eager:
                _task_stack.pop()
        return result

    def run_in_same_process(self, args, kw):
        try:
            return self.run(*args, **kw)
        except Abort as handle:
            transaction.abort()
            handle()
            return handle.message
        except HandleAfterAbort as handle:
            handle()
            raise

    def run_in_worker(self, principal_id, args, kw, retries=0):
        with self.configure_zope():
            try:
                with self.transaction(principal_id):
                    return self.run(*args, **kw)
            except HandleAfterAbort as handle:
                for handle_retries in range(self.max_retries):
                    try:
                        with self.transaction(principal_id):
                            handle()
                    except celery.exceptions.Retry:
                        # We have to handle ConflictErrors manually, since we
                        # don't want to retry the whole task (which was
                        # erroneous anyway), but only the after-abort portion.
                        countdown = random.uniform(0, 2 ** handle_retries)
                        log.warning('Waiting %s seconds for retry.', countdown)
                        time.sleep(countdown)
                        continue
                    else:
                        break
                else:
                    log.warning('Giving up on %r after %s retries',
                                handle, self.max_retries)

                if isinstance(handle, Abort):
                    return handle.message
                else:
                    raise handle

    @contextlib.contextmanager
    def transaction(self, principal_id):
        if principal_id:
            transaction.begin()
            login_principal(get_principal(principal_id), self.name)
            txn = transaction.get()
            txn.setUser(str(principal_id))
            txn.setExtendedInfo('task_name', self.name)
        try:
            yield
        except Exception:
            transaction.abort()
            raise
        else:
            try:
                transaction.commit()
            except ZODB.POSException.ConflictError:
                log.warning('Conflict while publishing', exc_info=True)
                transaction.abort()
                self.retry(
                    countdown=random.uniform(0, 2 ** self.request.retries))
        finally:
            transaction.abort()
            zope.security.management.endInteraction()

    @contextlib.contextmanager
    def configure_zope(self):
        old_site = zope.component.hooks.getSite()
        db = self.app.conf.get('ZODB')
        connection = db.open()
        root_folder = connection.root()[
            zope.app.publication.zopepublication.ZopePublication.root_name]
        zope.component.hooks.setSite(root_folder)
        try:
            yield
        finally:
            connection.close()
            zope.component.hooks.setSite(old_site)

    _eager_use_session_ = False  # Hook for tests

    def apply_async(self, args=None, kw=None, task_id=None, **options):
        self._assert_json_serializable(args, kw)
        if kw is None:
            kw = {}
        if task_id is None:
            task_id = celery.utils.gen_unique_id()
        kw.setdefault('_principal_id_', self._get_current_principal_id())

        # Accomodations for tests:
        # 1. Normally we defer (asynchronous) task execution to the transaction
        # commit via celery_session (see else-branch below). But in
        # always_eager mode the execution is synchronous, so the whole task
        # would run _during_ the commit phase, which totally breaks if inside
        # the task the transaction is touched in any way, e.g. by calling for a
        # savepoint. Since we need to support this kind of completely normal
        # behaviour, we bypass the session in always_eager mode.
        # 2. To simplify tests concerning our celery-integration mechanics we
        # provide a hook so tests can force using the session even in
        # always_eager mode (because otherwise those tests would have to use an
        # end-to-end setup, which would make the introspection they need
        # complicated if not impossible).
        # XXX These actually rather belongs into CelerySession, but that would
        # get mechanically complicated.
        if self.app.conf['task_always_eager'] and not self._eager_use_session_:
            self.__call__(*args, **kw)
        elif self.name == 'celery.ping':
            # Part of celery.contrib.testing setup, we need to perform this
            # immediately, because it has no transaction integration.
            return super().apply_async(args, kw, task_id=task_id, **options)
        else:
            # Hook so tests can force __call__ to use run_in_worker even when
            # always_eager is True, by passing in this kw explicitly.
            kw.setdefault('_run_asynchronously_', True)
            celery_session.add_call(
                super().apply_async, args, kw, task_id, **options)
        return self.AsyncResult(task_id)

    def _assert_json_serializable(self, *args, **kw):
        json.dumps(args)
        json.dumps(kw)

    def _get_current_principal_id(self):
        interaction = zope.security.management.queryInteraction()
        if interaction is None:
            principal_id = None
        else:
            principal_id = interaction.participations[0].principal.id
        return principal_id

    def retry(self, args=None, kwargs=None, exc=None, throw=True,
              eta=None, countdown=None, max_retries=None, **options):
        # copy&paste from superclass, we need to use a Retry exception that
        # uses the HandleAfterAbort mechanics to integrate with transactions.
        request = self.request
        retries = request.retries + 1
        max_retries = self.max_retries if max_retries is None else max_retries

        if request.called_directly:
            raise_with_context(
                exc or celery.exceptions.Retry('Task can be retried', None))

        if not eta and countdown is None:
            countdown = self.default_retry_delay

        # Add interaction information now, while it's still present. Our
        # apply_async() would normally do this, but it's called via
        # HandleAfterAbort when the original interaction has already ended.
        options.setdefault('_principal_id_', self._get_current_principal_id())

        S = self.signature_from_request(
            request, args, kwargs,
            countdown=countdown, eta=eta, retries=retries,
            **options
        )

        if max_retries is not None and retries > max_retries:
            if exc:
                raise_with_context(exc)
            raise self.MaxRetriesExceededError(
                "Can't retry {0}[{1}] args:{2} kwargs:{3}".format(
                    self.name, request.id, S.args, S.kwargs))

        ret = Retry(exc=exc, when=eta or countdown, signature=S)

        if request.is_eager:
            S.apply().get()
            if throw:
                raise ret
            return ret

        if throw:
            raise ret
        else:
            ret()
            return ret


CELERY = celery.Celery(
    __name__, task_cls=TransactionAwareTask, loader=ZopeLoader,
    strict_typing=False)



from ..logging import TaskFormatter
from .shared_tasks import shared_logging_task
import logging
import pytest
import tempfile
import transaction
import uuid


@pytest.fixture(scope='function')
def logger_and_stream():
    """Return a logger and the stream it writes to."""
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    from io import StringIO
    logged = StringIO()
    handler = logging.StreamHandler(logged)
    handler.setFormatter(
        TaskFormatter('task_id: %(task_id)s name: %(task_name)s %(message)s'))
    logger.addHandler(handler)

    yield (logger, logged)

    logger.removeHandler(handler)


def test_logging__TaskFormatter__format__1(logger_and_stream):
    """It provides task_id and task_name, which are empty if there is none."""

    log, logged = logger_and_stream
    log.debug("We have no task.")

    assert 'task_id:  name:  We have no task.\n' == logged.getvalue()


def test_logging__TaskFormatter__format__2(celery_session_worker):
    """It provides task_id and task_name in async mode."""

    task_id = '<task_id_{}>'.format(uuid.uuid4())
    with tempfile.NamedTemporaryFile(delete=False) as logfile:
        result = shared_logging_task.apply_async(
            (logfile.name,), task_id=task_id)
        transaction.commit()

        assert "Successful logged." == result.get()
        logfile.seek(0)
        log_result = logfile.read().decode('utf-8')
        assert ('task_id: {} '
                'name: z3c.celery.tests.shared_tasks.shared_logging_task '
                'we are logging'.format(task_id) in log_result)
        assert "__traceback_info__: we can handle traceback info" in log_result
        assert "NotImplementedError" in log_result



import celery.contrib.testing.app
import collections
import contextlib
import os
import pkg_resources
import plone.testing.zca
import pytest
import tempfile
import transaction
import z3c.celery
import z3c.celery.celery
import z3c.celery.testing
import zope.principalregistry.principalregistry
import zope.security.management
import zope.security.testing


ZODBConnection = collections.namedtuple(
    'ZODBConnection', ['connection', 'rootFolder', 'zodb'])


@pytest.fixture(scope='function')
def zcml():
    """Load ZCML on session scope."""
    layer = plone.testing.zca.ZCMLSandbox(
        name="CeleryZCML", filename='ftesting.zcml',
        module=__name__, package=z3c.celery)
    layer.setUp()
    yield layer
    # We might define principles in ftesting.zcml and we want to have a clean
    # state for each test, so we clear the registry here.
    layer.tearDown()
    zope.principalregistry.principalregistry.principalRegistry._clear()


@pytest.fixture(scope='function', autouse=True)
def automatic_transaction_begin():
    """Starts a new transaction for every test.

    We want to start with an empty celery_session for each test.

    """
    transaction.begin()
    zope.security.management.endInteraction()


@pytest.fixture(scope='function')
def interaction(automatic_transaction_begin):
    """Provide a zope interaction per test. Yields the principal."""
    principal = zope.security.testing.Principal(
        u'zope.user',
        groups=['zope.Authenticated'],
        description=u'test@example.com')
    z3c.celery.celery.login_principal(principal)
    yield principal
    zope.security.management.endInteraction()


@pytest.fixture(scope='session')
def storage_file():
    with tempfile.NamedTemporaryFile() as storage_file:
        yield storage_file.name


@pytest.fixture(scope='session')
def zope_conf(storage_file):
    with _zope_conf(storage_file) as x:
        yield x


@contextlib.contextmanager
def _zope_conf(storage_file):
    with tempfile.NamedTemporaryFile() as conf:
        conf.write(
            z3c.celery.testing.ZOPE_CONF_TEMPLATE.format(
                zodb_path=storage_file,
                ftesting_path=pkg_resources.resource_filename(
                    'z3c.celery', 'ftesting.zcml'),
                product_config='').encode('ascii'))
        conf.flush()
        yield conf.name


@pytest.fixture(scope='function')
def eager_celery_app(zope_conf):
    app = z3c.celery.CELERY
    conf = app.conf
    # deepcopy fails on py3 with infinite recursion,
    # but tests pass also with a simple copy of the conf
    old_conf = dict(conf)
    conf['ZOPE_CONF'] = zope_conf
    conf['task_always_eager'] = True
    conf['task_eager_propagates'] = True
    with celery.contrib.testing.app.setup_default_app(app):
        app.set_current()
        yield app
    app.conf.update(old_conf)


@pytest.fixture(scope='session')
def celery_config(zope_conf):
    return _celery_config(zope_conf)


def _celery_config(zope_conf):
    return {
        'broker_url': os.environ['Z3C_CELERY_BROKER'],
        'result_backend': os.environ['Z3C_CELERY_BROKER'],
        'worker_send_task_events': True,
        'task_send_sent_event': True,
        'task_remote_tracebacks': True,
        'ZOPE_CONF': zope_conf,
    }


@pytest.fixture(scope='session')
def celery_parameters():
    return _celery_parameters()


def _celery_parameters():
    return {
        'task_cls': z3c.celery.celery.TransactionAwareTask,
        'strict_typing': False,
        'loader': z3c.celery.celery.ZopeLoader,
    }


@pytest.fixture(scope='session')
def celery_enable_logging():
    return True


@pytest.fixture(scope='session')
def celery_includes():
    return ('z3c.celery.tests.shared_tasks',)


@pytest.fixture(scope='session')
def celery_worker_pool():
    return 'prefork'



# coding: utf-8
from .shared_tasks import get_principal_title_task
from celery import shared_task
from z3c.celery.celery import HandleAfterAbort, Abort
from z3c.celery.session import celery_session
from z3c.celery.testing import open_zodb_copy
import ZODB.POSException
import datetime
from unittest import mock
import pytest
import transaction
import z3c.celery
import z3c.celery.celery
import z3c.celery.testing
import zope.authentication.interfaces
import zope.security.management


@z3c.celery.task
def eager_task(context=None, datetime=None):
    """Dummy task to be used together with `eager_celery_app`."""


now = datetime.datetime.now()


def test_celery__TransactionAwareTask__delay__1():
    """It raises a TypeError if arguments are not JSON serializable."""
    with pytest.raises(TypeError):
        eager_task.delay(object(), datetime=now)


def test_celery__TransactionAwareTask__delay__2(interaction):
    """It registers a task if arguments are JSON serializable."""
    assert 0 == len(celery_session)
    eager_task.delay('http://url.to/testcontent',
                     datetime='2016-01-01 12:00:00')
    assert 1 == len(celery_session)


def test_celery__TransactionAwareTask__delay__3(interaction, eager_celery_app):
    """It extracts the principal from the interaction if run in async mode."""
    asynch = 'z3c.celery.celery.TransactionAwareTask._eager_use_session_'
    with mock.patch(asynch, new=True):
        eager_task.delay('1st param', datetime='now()')
    task_call = 'z3c.celery.celery.TransactionAwareTask.__call__'
    with mock.patch(task_call) as task_call:
        zope.security.management.endInteraction()
        transaction.commit()
    task_call.assert_called_with(
        '1st param', datetime='now()',
        _run_asynchronously_=True, _principal_id_=u'zope.user')


def test_celery__TransactionAwareTask__delay__4(interaction, eager_celery_app):
    """It calls the original function directly if we are in eager mode."""
    with mock.patch.object(eager_task, 'run') as task_call:
        eager_task.delay('1st param', datetime='now()')
    task_call.assert_called_with('1st param', datetime='now()')


def test_celery__TransactionAwareTask__delay__5(celery_session_worker, zcml):
    """It allows to run two tasks in a single session."""
    auth = zope.component.getUtility(
        zope.authentication.interfaces.IAuthentication)
    principal = auth.getPrincipal('example.user')
    z3c.celery.celery.login_principal(principal)
    result1 = get_principal_title_task.delay()

    zope.security.management.endInteraction()
    principal = auth.getPrincipal('zope.user')
    z3c.celery.celery.login_principal(principal)
    result2 = get_principal_title_task.delay()

    transaction.commit()

    assert 'Ben Utzer' == result1.get()
    assert 'User' == result2.get()


def test_celery__TransactionAwareTask__delay__6(celery_session_worker, zcml):
    """It allows overriding the principal."""
    auth = zope.component.getUtility(
        zope.authentication.interfaces.IAuthentication)
    principal = auth.getPrincipal('example.user')
    z3c.celery.celery.login_principal(principal)
    result = get_principal_title_task.delay(_principal_id_='zope.user')

    transaction.commit()
    assert 'User' == result.get()


def test_celery__TransactionAwareTask__apply_async__1():
    """It raises a TypeError if arguments are not JSON serializable."""
    with pytest.raises(TypeError):
        eager_task.apply_async(
            (object(),), {'datetime': now},
            task_id=now, countdown=now)


def test_celery__TransactionAwareTask__apply_async__2(interaction):
    """It registers a task if arguments are JSON serializable."""
    assert 0 == len(celery_session)
    eager_task.apply_async(
        ('http://url.to/testcontent',),
        {'datetime': '2016-01-01 12:00:00'},
        task_id=1, countdown=30)
    assert 1 == len(celery_session)


def test_celery__TransactionAwareTask__apply_async__3(
        interaction, eager_celery_app):
    """It extracts the principal from the interaction if run in async mode."""
    asynch = 'z3c.celery.celery.TransactionAwareTask._eager_use_session_'
    with mock.patch(asynch, new=True):
        eager_task.apply_async(('1st param',), dict(datetime='now()'))
    task_call = 'z3c.celery.celery.TransactionAwareTask.__call__'
    with mock.patch(task_call) as task_call:
        zope.security.management.endInteraction()
        transaction.commit()
    task_call.assert_called_with(
        '1st param', datetime='now()',
        _run_asynchronously_=True, _principal_id_=u'zope.user')


def test_celery__TransactionAwareTask__apply_async__4(
        interaction, eager_celery_app):
    """It calls the original function directly if we are in eager mode."""
    with mock.patch.object(eager_task, 'run') as task_call:
        eager_task.apply_async(('1st param',), dict(datetime='now()'))
    task_call.assert_called_with('1st param', datetime='now()')


def test_celery__TransactionAwareTask__apply_async__5(
        interaction, eager_celery_app):
    """It processes the missing kw argument as emtpy dict."""
    # We need to make this a bit complicated as we want to see the actual call
    # of __call__ after the argument handling
    asynch = 'z3c.celery.celery.TransactionAwareTask._eager_use_session_'
    with mock.patch(asynch, new=True):
        eager_task.apply_async(('1st param',))
    task_call = 'z3c.celery.celery.TransactionAwareTask.__call__'
    with mock.patch(task_call) as task_call:
        transaction.commit()
    task_call.assert_called_with(
        '1st param', _run_asynchronously_=True, _principal_id_=u'zope.user')


@shared_task
def exception_task(context=None, datetime=None):
    """Dummy task which raises an exception to test our framework."""
    raise RuntimeError()


def test_celery__TransactionAwareTask____call____1(celery_session_worker):
    """It aborts the transaction in case of an error during task execution."""
    result = exception_task.delay()
    transaction.commit()
    with pytest.raises(Exception) as err:
        result.get()
    # Celery wraps errors dynamically as celery.backends.base.<ErrorName>, so
    # we have to dig deep here.
    assert 'RuntimeError' == err.value.__class__.__name__


def test_celery__TransactionAwareTask____call____1__cov(
        interaction, eager_celery_app):
    """It aborts the transaction in case of an error during task execution.

    As it is hard to collect coverage for subprocesses we use this test for
    coverage only.
    """
    configure_zope = 'z3c.celery.celery.TransactionAwareTask.configure_zope'
    with mock.patch(configure_zope), \
            mock.patch.object(eager_task, 'run',
                              side_effect=RuntimeError) as run, \
            mock.patch('transaction.abort') as abort:

        zope.security.management.endInteraction()
        with pytest.raises(RuntimeError):
            # We want to simulate a run in worker. The RuntimeError is raised
            # by the mock
            eager_task(_run_asynchronously_=True)

    assert run.called
    assert abort.called


@shared_task(max_retries=1)
def conflict_task(bind=True, context=None, datetime=None):
    """Dummy task which injects a DataManager that votes a ConflictError."""
    transaction.get().join(VoteExceptionDataManager())


class NoopDatamanager:
    """Datamanager which does nothing."""

    def abort(self, trans):
        pass

    def commit(self, trans):
        pass

    def tpc_begin(self, trans):
        pass

    def tpc_abort(self, trans):
        pass


class VoteExceptionDataManager(NoopDatamanager):
    """DataManager which raises an exception in tpc_vote."""

    def tpc_vote(self, trans):
        raise ZODB.POSException.ConflictError()

    def sortKey(self):
        # Make sure we are running after the in thread execution of the task so
        # that we can throw an Error as last part of vote:
        return '~~sort-me-last'


def test_celery__TransactionAwareTask____call____2(
        celery_session_worker, interaction):
    """It aborts the transaction and retries in case of an ConflictError."""
    result = conflict_task.delay()
    transaction.commit()
    with pytest.raises(Exception) as err:
        result.get()
    assert 'MaxRetriesExceededError' == err.value.__class__.__name__


@shared_task(max_retries=2)
def conflict_after_abort_task():
    """Dummy task which injects a DataManager that votes a ConflictError."""
    raise HandleAfterAbort(conflicts_once)


conflicts_after_abort = 0


def conflicts_once():
    global conflicts_after_abort
    if conflicts_after_abort == 0:
        transaction.get().join(VoteExceptionDataManager())
    conflicts_after_abort += 1


def test_celery__TransactionAwareTask____call____2a(eager_celery_app):
    """It retries on ConflictError during HandleAfterAbort."""
    assert conflicts_after_abort == 0
    configure_zope = 'z3c.celery.celery.TransactionAwareTask.configure_zope'
    with mock.patch(configure_zope), \
            mock.patch('time.sleep') as sleep:
        with pytest.raises(HandleAfterAbort):
            conflict_after_abort_task(_run_asynchronously_=True)
        assert sleep.call_count == 1
        assert conflicts_after_abort == 2


def test_celery__TransactionAwareTask____call____3(
        celery_session_worker, zcml):
    """It runs as given principal in asynchronous mode."""
    auth = zope.component.getUtility(
        zope.authentication.interfaces.IAuthentication)
    principal = auth.getPrincipal('example.user')
    z3c.celery.celery.login_principal(principal)
    result = get_principal_title_task.delay()
    transaction.commit()
    assert 'Ben Utzer' == result.get()


def test_celery__TransactionAwareTask____call____3__cov(
        interaction, eager_celery_app, zcml):
    """It runs as given principal in asynchronous mode.

    As it is hard to collect coverage for sub-processes we use this test for
    coverage only.
    """
    configure_zope = 'z3c.celery.celery.TransactionAwareTask.configure_zope'
    with mock.patch(configure_zope):

        zope.security.management.endInteraction()
        result = get_principal_title_task(
            _run_asynchronously_=True, _principal_id_='example.user')

    assert "Ben Utzer" == result


@shared_task(bind=True)
def get_task_id(self):
    """Get the task id of the job."""
    return self.request.id


def test_celery__TransactionAwareTask____call____4(
        celery_session_worker, interaction):
    """It propagates the task_id to the worker."""
    job = get_task_id.apply_async(task_id='my-nice-task-id')
    transaction.commit()
    assert 'my-nice-task-id' == job.get()


@shared_task
def except_with_handler():
    """Raise an exception which is handled after abort."""
    site = zope.component.hooks.getSite()
    site['foo'] = 'bar'

    def handler(arg1, arg2, kw1=1, kw2=2):
        interaction = zope.security.management.getInteraction()
        site['data'] = (arg1, arg2, kw1, kw2,
                        interaction.participations[0].principal.title)

    raise HandleAfterAbort(handler, 'a1', 'a2', kw2=4)


def test_celery__TransactionAwareTask____call__5(
        interaction):
    """It handles specific exceptions when running synchronously."""
    data = {}
    with mock.patch('zope.component.hooks.getSite', return_value=data):
        with pytest.raises(HandleAfterAbort):
            except_with_handler()
    assert {'data': ('a1', 'a2', 1, 4, u'zope.user'), 'foo': 'bar'} == data


def test_celery__TransactionAwareTask__run_in_worker__1(
        celery_session_worker, storage_file, interaction):
    """It handles specific exceptions in a new transaction after abort."""
    job = except_with_handler.delay()
    transaction.commit()
    with pytest.raises(Exception):
        job.get()

    with open_zodb_copy(storage_file) as app:
        assert [('data', ('a1', 'a2', 1, 4, u'User'))] == list(app.items())


def test_celery__TransactionAwareTask__run_in_worker__1__cov(
        interaction, eager_celery_app, zcml):
    """It handles specific exceptions in a new transaction after abort.

    As it is hard to collect coverage for sub-processes we use this test for
    coverage only.
    """
    data = {}
    configure_zope = 'z3c.celery.celery.TransactionAwareTask.configure_zope'
    with mock.patch(configure_zope),\
            mock.patch('zope.component.hooks.getSite', return_value=data):
        zope.security.management.endInteraction()
        with pytest.raises(HandleAfterAbort):
            except_with_handler(
                _run_asynchronously_=True, _principal_id_='example.user')

    # transaction.abort() does not remove items from a dict, so 'foo': 'bar'
    # also shows up here:
    assert {'data': ('a1', 'a2', 1, 4, u'Ben Utzer'), 'foo': 'bar'} == data


@shared_task
def success_but_abort_transaction():
    site = zope.component.hooks.getSite()
    site['flub'] = 'will not be written'
    raise Abort(lambda: None, message='done')


def test_celery__TransactionAwareTask__run_in_worker__2(
        celery_session_worker, storage_file, interaction):
    """It handles a specific exceptions to abort the transaction but still
    count as a successful job."""
    job = success_but_abort_transaction.delay()
    transaction.commit()
    assert 'done' == job.get()

    with open_zodb_copy(storage_file) as app:
        assert 'flub' not in app


def test_celery__HandleAfterAbort__1():
    """It returns the message in the exception unicode which was passed in."""
    err = HandleAfterAbort(lambda: None, message=u'test-messäge')
    err()
    assert 'test-messäge' == str(err)


def test_celery__HandleAfterAbort__1_5():
    """It returns the message in the exception string which was passed in."""
    err = HandleAfterAbort(lambda: None, message='test-messäge')
    err()
    assert 'test-messäge' == str(err)


def test_celery__HandleAfterAbort__2():
    """It does not pass the message as argument to the callback."""
    data = {}

    def save_args(*args, **kwargs):
        data['args'] = args
        data['kwargs'] = kwargs

    err = HandleAfterAbort(save_args, 2, 3, message=u'test-messäge', kw1='23')
    err()
    assert {'args': (2, 3), 'kwargs': {'kw1': '23'}} == data


class CommitExceptionDataManager(NoopDatamanager):
    """DataManager which raises an exception in tpc_vote."""

    def tpc_vote(self, trans):
        raise RuntimeError('provoked')

    def sortKey(self):
        # Make sure we are running after the in thread execution of the task so
        # that we can throw an Error as last part of vote:
        return '~~sort-me-last'


@shared_task
def commit_error_task(bind=True):
    transaction.get().join(CommitExceptionDataManager())


def test_celery__TransactionAwareTask__run_in_worker__3(
        celery_session_worker, interaction):
    """It always ends the interaction even when commit raises."""
    commit_error_task.delay()
    transaction.commit()
    result = get_principal_title_task.delay()
    transaction.commit()
    assert 'User' == result.get()


@shared_task(bind=True)
def get_principal_retry(self):
    if not self.request.retries:
        self.retry(countdown=0)
    interaction = zope.security.management.getInteraction()
    return [
        self.request.retries, interaction.participations[0].principal.title]


def test_celery__TransactionAwareTask_retry(
        interaction, celery_session_worker):
    """It keeps the interaction on retry."""
    result = get_principal_retry.delay()
    transaction.commit()
    assert [1, 'User'] == result.get()



from celery import shared_task
import zope.security.management
import logging
from ..logging import TaskFormatter


@shared_task
def get_principal_title_task():
    """Task returning the principal's title used to run it."""
    interaction = zope.security.management.getInteraction()
    return interaction.participations[0].principal.title


def log_exception(filename):
    """Common part for logging tasks."""
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)
    handler = logging.FileHandler(filename)
    handler.setFormatter(
        TaskFormatter('task_id: %(task_id)s name: %(task_name)s %(message)s'))
    log.addHandler(handler)

    __traceback_info__ = ('we can handle traceback info')
    try:
        raise NotImplementedError()
    except NotImplementedError:
        log.error('we are logging', exc_info=True)
    handler.flush()
    handler.close()
    return "Successful logged."


@shared_task(bind=True)
def shared_logging_task(self, filename):
    return log_exception(filename)



import ZODB
import ZODB.FileStorage
import contextlib
import os
import shutil


ZOPE_CONF_TEMPLATE = """
site-definition {ftesting_path}

<zodb>
  <filestorage>
    path {zodb_path}
  </filestorage>
</zodb>

{product_config}

# eventlog is required :-(
<eventlog>
</eventlog>
"""


@contextlib.contextmanager
def open_zodb_copy(zodb_path):
    """Context manager which opens a copy of the given ZODB.

    This might be useful if the actual ZODB is still opened by another process.
    Yields the application object inside the ZODB.
    """
    new_zodb_path = zodb_path + '.copy'
    shutil.copy(zodb_path, new_zodb_path)
    zodbDB = ZODB.DB(ZODB.FileStorage.FileStorage(file_name=new_zodb_path))
    connection = zodbDB.open()
    try:
        yield connection.root()['Application']
    finally:
        connection.close()
        zodbDB.close()
        os.unlink(new_zodb_path)



from ..layer import EndToEndLayer
from .shared_tasks import get_principal_title_task
import plone.testing
import plone.testing.zca
import tempfile
import transaction
import unittest
import z3c.celery.celery
import z3c.celery.conftest
import z3c.celery.testing
import zope.authentication.interfaces
import zope.component


ZCML_LAYER = plone.testing.zca.ZCMLSandbox(
    name="CeleryZCML", filename='ftesting.zcml',
    module=__name__, package=z3c.celery)


class SettingsLayer(plone.testing.Layer):
    """Settings for the EndToEndLayer."""

    def setUp(self):
        self['storage_file_fixture'] = tempfile.NamedTemporaryFile()
        storage_file = self['storage_file_fixture'].__enter__()
        self['zope_conf_fixture'] = z3c.celery.conftest._zope_conf(
            storage_file.name)

        zope_conf = self['zope_conf_fixture'].__enter__()
        self['celery_config'] = z3c.celery.conftest._celery_config(zope_conf)
        self['celery_parameters'] = z3c.celery.conftest._celery_parameters()
        self['celery_worker_parameters'] = {'queues': ('hiprio', 'celery')}
        self['celery_includes'] = ['z3c.celery.tests.shared_tasks']

    def tearDown(self):
        self['zope_conf_fixture'].__exit__(None, None, None)
        del self['zope_conf_fixture']
        self['storage_file_fixture'].__exit__(None, None, None)
        del self['storage_file_fixture']
        del self['celery_config']
        del self['celery_includes']
        del self['celery_parameters']
        del self['celery_worker_parameters']


SETTINGS_LAYER = SettingsLayer()
CONFIGURED_END_TO_END_LAYER = EndToEndLayer(
    bases=[SETTINGS_LAYER], name="ConfiguredEndToEndLayer")
ZOPE_END_TO_END_LAYER = plone.testing.Layer(
    bases=(CONFIGURED_END_TO_END_LAYER, ZCML_LAYER),
    name="ZopeEndToEndLayer")


class EndToEndLayerTests(unittest.TestCase):
    """Testing ..layer.EndToEndLayer."""

    layer = ZOPE_END_TO_END_LAYER

    def test_layer__EndToEndLayer__1(self):
        auth = zope.component.getUtility(
            zope.authentication.interfaces.IAuthentication)
        principal = auth.getPrincipal('example.user')
        z3c.celery.celery.login_principal(principal)
        result = get_principal_title_task.apply_async(queue='hiprio')

        transaction.commit()

        assert 'Ben Utzer' == result.get(timeout=10)



from celery.contrib.testing.app import TestApp, setup_default_app
from celery.contrib.testing.worker import start_worker
import plone.testing
import z3c.celery


class EagerLayer(plone.testing.Layer):

    def setUp(self):
        # No isolation problem, end to end tests use a separate celery app
        # which is provided by EndToEndLayer (below).
        z3c.celery.CELERY.conf.task_always_eager = True

    def tearDown(self):
        z3c.celery.CELERY.conf.task_always_eager = False


EAGER_LAYER = EagerLayer()


class EndToEndLayer(plone.testing.Layer):
    """Run celery end to end tests in a plone.testing Layer.

    Expects the following resources:

    * `celery_config`: dict of config options for the celery app
    * `celery_parameters`: dict of parameters used to instantiate Celery
    * `celery_worker_parameters`: dict of parameters used to instantiate
       celery workers
    * `celery_includes`: list of dotted names to load the tasks in the worker
    """

    def setUp(self):
        celery_app = TestApp(
            set_as_current=False, enable_logging=True,
            config=self['celery_config'], **self['celery_parameters'])
        self['celery_app_fixture'] = setup_default_app(celery_app)
        self['celery_app_fixture'].__enter__()

        for module in self['celery_includes']:
            celery_app.loader.import_task_module(module)

        self['celery_worker_fixture'] = start_worker(
            celery_app, pool='prefork', **self['celery_worker_parameters'])
        self['celery_worker_fixture'].__enter__()

    def tearDown(self):
        self['celery_app_fixture'].__exit__(None, None, None)
        del self['celery_app_fixture']
        self['celery_worker_fixture'].__exit__(None, None, None)
        del self['celery_worker_fixture']



import celery.concurrency.asynpool
import celery.loaders.app
import celery.signals
import celery.utils.collections
import logging.config
import types
import os.path
import zope.app.wsgi


class ZopeLoader(celery.loaders.app.AppLoader):
    """Sets up the Zope environment in the Worker processes."""

    def on_worker_init(self):
        logging_ini = self.app.conf.get('LOGGING_INI')
        if not logging_ini:
            return

        @celery.signals.setup_logging.connect(weak=False)
        def setup_logging(*args, **kw):
            """Make the loglevel finely configurable via a config file."""
            config_file = os.path.abspath(logging_ini)
            logging.config.fileConfig(
                config_file, dict(
                    __file__=config_file, here=os.path.dirname(config_file)),
                disable_existing_loggers=False)

        if self.app.conf.get('DEBUG_WORKER'):
            assert self.app.conf.get('worker_pool') == 'solo'
            self.on_worker_process_init()

        # Work around <https://github.com/celery/celery/issues/4323>.
        if self.app.conf.get('worker_boot_timeout'):
            celery.concurrency.asynpool.PROC_ALIVE_TIMEOUT = float(
                self.app.conf['worker_boot_timeout'])

    def on_worker_process_init(self):
        conf = self.app.conf
        configfile = conf.get('ZOPE_CONF')
        if not configfile:
            raise ValueError(
                'Celery setting ZOPE_CONF not set, '
                'check celery worker config.')

        db = zope.app.wsgi.config(configfile)
        conf['ZODB'] = db

    def on_worker_shutdown(self):
        if 'ZODB' in self.app.conf:
            self.app.conf['ZODB'].close()

    def read_configuration(self):
        """Read configuration from either

        * an importable python module, given by its dotted name in
          CELERY_CONFIG_MODULE. Note that this can also be set via
          `$ bin/celery worker --config=<modulename>`. (Also note that "celery
          worker" includes the cwd on the pythonpath.)
        * or a plain python file (given by an absolute path in
          CELERY_CONFIG_FILE)

        If neither env variable is present, no configuration is read, and some
        defaults are used instead that most probably don't work (they assume
        amqp on localhost as broker, for example).
        """
        module = os.environ.get('CELERY_CONFIG_MODULE')
        if module:
            return super().read_configuration()
        pyfile = os.environ.get('CELERY_CONFIG_FILE')
        if pyfile:
            module = self._import_pyfile(pyfile)
            return celery.utils.collections.DictAttribute(module)

    def _import_pyfile(self, filename):
        """Applies Celery configuration by reading the given python file
        (absolute filename), which unfortunately Celery does not support.

        (Code inspired by flask.config.Config.from_pyfile)
        """
        module = types.ModuleType('config')
        module.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(
                    config_file.read(), filename, 'exec'), module.__dict__)
        except IOError as e:
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise e
        else:
            return module



import datetime
import os
import os.path
import shutil


author = 'gocept, Zeit Online'
_year_started = 2016
_year = datetime.date.today().year
if _year != _year_started:
    _year = u'%s-%s' % (_year_started, _year)
copyright = u' %s %s' % (_year, author)
source_suffix = '.rst'
master_doc = 'index'

needs_sphinx = '1.0'
extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
]

templates_path = [
    '.',
]

html_sidebars = {
    '**': ['globaltoc.html', 'searchbox.html', 'project-links.html'],
}


# We use the autosummary extension to build API docs from source code.
# However, this extension doesn't update the generated docs if the source
# files change. Therefore, we need to remove the generated stuff before
# each run. The _autosummary_output variable tells the relative path to
# the directory that autosummary uses to put its generated files and which
# we, therefore, need to remove. It must be the same that the autosummary
# directive in api.rst points to.

autosummary_generate = ['api.rst']
_autosummary_output = './_api/'
if os.path.isdir(_autosummary_output):
    shutil.rmtree(_autosummary_output)
elif os.path.exists(_autosummary_output):
    raise RuntimeError('Expected %s to be a directory.' %
                       os.path.abspath(_autosummary_output))
os.mkdir(_autosummary_output)



# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""Integration of Celery 4 with Zope 3.
"""

from setuptools import setup, find_packages

setup(
    name='z3c.celery',
    version='1.7.0',

    install_requires=[
        'celery >= 4.0.2',
        'setuptools',
        'transaction',
        'ZODB',
        'zope.app.appsetup',
        'zope.app.publication',
        'zope.app.wsgi',
        'zope.authentication',
        'zope.component',
        'zope.exceptions',
        'zope.interface',
        'zope.publisher',
        'zope.security',
        'zope.principalregistry',
    ],

    extras_require={
        'test': [
            'pytest',
            'pytest-celery',
            'gocept.pytestlayer',
            'plone.testing',
            'redis',
            'tblib',
            'zope.traversing',
        ],
        'layer': [  # use this extra when using the EndToEndLayer
            'plone.testing',
        ],

    },

    entry_points={
        'console_scripts': [
            # 'binary-name = z3c.celery.module:function'
        ],
    },

    author='gocept, Zeit Online',
    author_email='zon-backend@zeit.de',
    license='BSD',
    url='https://github.com/ZeitOnline/z3c.celery',
    keywords='celery Zope transaction',
    classifiers="""\
Development Status :: 3 - Alpha
Framework :: Zope3
Intended Audience :: Developers
License :: OSI Approved
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python
Programming Language :: Python :: 3.7
Programming Language :: Python :: Implementation :: CPython
Topic :: Database
Topic :: Software Development
Topic :: Software Development :: Testing
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(open(name).read() for name in (
        'README.rst',
        'CHANGES.rst',
    )),
    namespace_packages=['z3c'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
)
